/*  martss.go - MARTY Real-time Scheduling Service written in Go
 *    Copyright (C) 2023-2024  hoff.industries
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <https://www.gnu.org/licenses/>. */

package main

import (
	"encoding/hex"
	"fmt"
	"os"
	"time"

	"hoff.industries/marty/libmarty"
	"hoff.industries/marty/libmmp"
	"hoff.industries/marty/martcconf"
	"hoff.industries/marty/mendpoints"
)

var TARGET_PROTOCOL = []int{2, 0, 0}
var INSTANCE_VERSION = []int{3, 1, 3}

const INSTANCE_NAME string = "m-gomartss"
const INSTANCE_TITLE string = "MaRTSS - Real-Time Scheduling Service"

var Endpoints = []libmarty.MartyClientEndpoint{

	{
		Endpoint: mendpoints.MARTSS_EPT_EVENT,
		Name: "event",
		// oTime, oTitle, Oneshot, DayMask, MonthMask, Skip
		Min: 11,
		Max: 11,
		IsAscii: true,
	},

	{
		Endpoint: mendpoints.MARTSS_EPT_APPEND,
		Name: "append",
		// oTime, oTitle, Payloads...
		Min: 6,
		Max: -1,
		IsAscii: true,
	},

	{
		Endpoint: mendpoints.MARTSS_EPT_SKIP,
		Name: "skip",
		// oTime, oTitle
		Min: 7,
		Max: 7,
		IsAscii: true,
	},

	{
		Endpoint: mendpoints.MARTSS_EPT_RESCHEDULE,
		Name: "reschedule",
		// oTime, oTitle, nTime
		Min: 8,
		Max: 8,
		IsAscii: true,
	},
}

var MaxEndpoint = len(Endpoints)

func event_exists(title string, time string) bool{

	tickEvents, mapLookup := martcconf.MaRTCConf[time]
	if (!mapLookup) {
		return false
	}

	for i := 0; i < len(tickEvents); i++ {

		if (tickEvents[i].Title == title) {
			return true
		}
	}

	return false
}

func application_loop(data *libmarty.MartyClientData) {

	var currentTick time.Time = time.Now()

	var tickString string = currentTick.Format("15:04:05")
	var tickDiff int = 0

	var tickEvents []martcconf.RTCEvent
	var tickEvent *martcconf.RTCEvent
	var mapLookup bool

	var status libmmp.MMStatus

	var shiftDigit int

	var replaceEvents []martcconf.RTCEvent = nil

	for true {

		tickDiff = int(time.Now().Sub(currentTick).Seconds())

		for (tickDiff < 1) {
			// poll current time every 50 ms to see if a new second has elapsed
			time.Sleep(time.Duration(50) * time.Millisecond)
			tickDiff = int(time.Now().Sub(currentTick).Seconds())
		}

		if (tickDiff > 1) {
			// we've skipped a tick - shouldn't happen ever but this is there
			// to let me know just in case I ever need to investigate it

			libmarty.Marty_full_log("WARNING: TICK SKIP OCCURED!\n", libmarty.LOG_WARNING, data)
		}

		// grab our current tick from our last one plus the tick diff, which should be 1s
		currentTick = currentTick.Add(time.Second * time.Duration(tickDiff))
		tickString = currentTick.Format("15:04:05")

		libmarty.Marty_full_log(fmt.Sprintf("=== MaRTSS tick: %s ===\r", tickString), libmarty.LOG_INFO, data)

		tickEvents, mapLookup = martcconf.MaRTCConf[tickString]

		if (!mapLookup) {
			// no events for this tick, so we do nothing
			continue
		}

		replaceEvents = nil

		for i := 0; i < len(tickEvents); i++ {

			tickEvent = &tickEvents[i]

			if (!tickEvent.Oneshot) {

				replaceEvents = append(replaceEvents, *tickEvent)
			} else {

				libmarty.Marty_full_log("Executing oneshot '%s' at %s; this won't repeat.\n",
					libmarty.LOG_WARNING, data, tickEvent.Title, tickString)
			}

			if (tickEvent.Skip) {

				tickEvent.Skip = false
				libmarty.Marty_full_log("Skipping '%s' at %s this time...\n",
					libmarty.LOG_WARNING, data, tickEvent.Title, tickString)
				continue
			}

			shiftDigit = int(currentTick.Month() - 1)

			if (tickEvent.MonthMask & (1 << shiftDigit) == 0x0000) {
				continue
			}

			// because sunday in golang is day 0 for some dumb reason
			// could swap to a LUT if no hw division on target
			shiftDigit = ((int(currentTick.Weekday()) - 1) % 7 + 7) % 7

			if (tickEvent.DayMask & (1 << shiftDigit) == 0x00) {
				continue
			}

			// previously this has been fast enough to execute all within the one second tick
			// if I experience tick skip though might be worth having a dispatcher loop
			// where I put the events on a buffered channel for execution...

			for j := 0; j < len(tickEvent.Messages); j++ {

				libmarty.Marty_message_handler(&tickEvent.Messages[j], &status, data)
				if (status.Error != libmmp.NONE) {

					libmarty.Marty_full_log("Library error 0x%x occured when executing!\n",
						libmarty.LOG_ERROR, data, status.Error)
				}
			}
		}

		if (len(replaceEvents) == 0) {

			libmarty.Marty_full_log("No more events for %s, removing entry.\n",
				libmarty.LOG_WARNING, data, tickString)
			delete(martcconf.MaRTCConf, tickString)
		} else {

			martcconf.MaRTCConf[tickString] = replaceEvents
		}
	}
}

func main() {
	// define standard base structures for my marty clientship
	var data libmarty.MartyClientData
	var status libmmp.MMStatus

	var txPacket libmmp.MMPacket
	var rxPacket *libmmp.MMPacket = nil

	var deliveryHandle *libmarty.MartyMessageHandle

	var applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["SUCCESS"]

	var hostname string = ""
	var callError error = nil

	var endpointConversion int = -1
	var endpoint *libmarty.MartyClientEndpoint = nil

	var validationMap map[string]bool
	var mapLookup bool

	var tickEvent *martcconf.RTCEvent
	var tickEvents []martcconf.RTCEvent

	var oTime string
	var nTime string
	var oTitle string

	var hexDecode []byte = nil
	var decodeError error = nil

	libmarty.Marty_client_data_factory(&data)

	libmmp.Mmstatus_factory(&status)
	libmmp.Mmpacket_factory(&txPacket, true)

	// set up initial client settings
	data.VersionString = fmt.Sprintf("%s v%d.%d.%d", INSTANCE_NAME,
		INSTANCE_VERSION[0], INSTANCE_VERSION[1], INSTANCE_VERSION[2])

	fmt.Printf("%s[%s]%s Starting %s %s - %s\n",
		   libmarty.LoggerColours[libmarty.LOG_INFO], libmarty.MartyFlavour,
	    libmarty.ClearColour, libmarty.LIBMARTY_AUTHOR, INSTANCE_TITLE, data.VersionString)

	hostname, callError = os.Hostname()
	if (callError != nil) {

		os.Exit(libmarty.EXIT_ERR_HOSTNAME)
	}

	data.ConnType = libmarty.CONN_UNIX
	data.InitParameters.Auth = ""
	data.InitParameters.Priority = 0
	data.InitParameters.Host = hostname
	data.InitParameters.Name = "martss"

	if (libmarty.Spawn_marty_socket(&data) != libmmp.NONE) {

		libmarty.Marty_full_log("Socket spawn failure!\n",
					libmarty.LOG_CRITICAL, &data)
		os.Exit(libmarty.EXIT_ERR_SPAWN)
	}

	if (libmarty.Marty_init(&data) != libmmp.NONE) {

		libmarty.Marty_full_log("MMP initialisation failure!\n",
					libmarty.LOG_CRITICAL, &data)
		os.Exit(libmarty.EXIT_ERR_INIT)
	};

	libmarty.Marty_full_log("MMP initialisation success!\n", libmarty.LOG_STARTUP, &data)

	go libmarty.Marty_network_loop(&data)
	libmarty.Marty_full_log("Network loop started.\n", libmarty.LOG_STARTUP, &data)

	go application_loop(&data)
	libmarty.Marty_full_log("Application loop started.\n", libmarty.LOG_STARTUP, &data)

	for i := range martcconf.MaRTCConf {

		validationMap = make(map[string]bool)

		for j := 0; j < len(martcconf.MaRTCConf[i]); j++ {

			_, mapLookup = validationMap[martcconf.MaRTCConf[i][j].Title]

			if (mapLookup) {
				libmarty.Marty_full_log("WARNING: Duplicate title entry for ", libmarty.LOG_WARNING, &data)
				fmt.Printf("%s at %s.\n", martcconf.MaRTCConf[i][j].Title, i)
			}

			validationMap[martcconf.MaRTCConf[i][j].Title] = true
		}
	}

	for true {

		libmmp.Mmpacket_reset(&txPacket, true)
		applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["SUCCESS"]

		if (rxPacket != nil) {
			libmmp.Mmpacket_reset(rxPacket, true)
		}

		rxPacket = nil

		select {
			case deliveryHandle = <- data.DeliveryChannel:
				switch (deliveryHandle.Result) {

					case libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["SUCCESS"]:
						break;

					case libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["DELIVERY"]:

						libmarty.Marty_full_log("Failed to execute event: delivery failed!\n",
							libmarty.LOG_ERROR, &data)

						continue

					default:
						libmarty.Marty_full_log("Failed to execute event: application level error ",
							libmarty.LOG_ERROR, &data)
						fmt.Printf("%x\n", deliveryHandle.Result)

						continue

				}

				if deliveryHandle.Next == nil {

					libmarty.Marty_full_log("The M.A.R-T>Y. Real-Time Scheduling Service - Who says being on time isn't fashionable?\n",
						libmarty.LOG_SUCCESS, &data)
					continue
				}

				libmarty.Marty_message_handler(deliveryHandle.Next, &status, &data)
				if (status.Error != libmmp.NONE) {

					libmarty.Marty_full_log("Error executing follow-on sequence member!\n",
						libmarty.LOG_WARNING, &data)
				}

				continue

				// listen for new SEND messages from client loop on channel
			case rxPacket = <- data.MessageChannel:
		}

		if (rxPacket == nil) {
			libmarty.Marty_full_log("Received NULL pointer from ITC channel!\n",
				libmarty.LOG_WARNING, &data)
			continue
		}

		endpointConversion = int(rxPacket.Payloads[libmarty.PAYL_IDX_ENDPOINT][0])

		if (endpointConversion > MaxEndpoint || endpointConversion < 0) {

			libmarty.Marty_full_log("Requested endpoint outside my range!\n",
				libmarty.LOG_WARNING, &data)
			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["ENDPOINT"]

			goto JMP_TX_STATUS
		}

		endpoint = &Endpoints[endpointConversion]

		if (len(rxPacket.Payloads) < endpoint.Min) {
			libmarty.Marty_full_log("Request payload count less than expected!\n",
				libmarty.LOG_WARNING, &data)

			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["TOO_FEW"]

			goto JMP_TX_STATUS
		}

		if (endpoint.Max > 0 && len(rxPacket.Payloads) > endpoint.Max) {
			libmarty.Marty_full_log("Request payload count more than expected!\n",
				libmarty.LOG_WARNING, &data)

			applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["TOO_MANY"]

			goto JMP_TX_STATUS
		}

		oTime = libmmp.Payload_2_str(rxPacket.Payloads[5])
		oTitle = libmmp.Payload_2_str(rxPacket.Payloads[6])

		tickEvent = nil
		tickEvents, mapLookup = martcconf.MaRTCConf[oTime]

		if (mapLookup) {
			for i := 0; i < len(tickEvents); i++ {

				tickEvent = &tickEvents[i]

				if (tickEvent.Title == oTitle) {
					break
				}

				tickEvent = nil
			}
		}

		switch (endpointConversion) {
			case mendpoints.MARTSS_EPT_EVENT:

				if (!mapLookup) {

					martcconf.MaRTCConf[oTime] = nil
					libmarty.Marty_full_log("Made a new events entry at %s.\n",
						libmarty.LOG_WARNING, &data, oTime)
				}

				if event_exists(oTitle, oTime) {

					libmarty.Marty_full_log("Event '%s' at %s already exists! Won't schedule.\n",
						libmarty.LOG_ERROR, &data, oTitle, oTime)
					applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["EXECUTE"]
					goto JMP_TX_STATUS
				}

				newEvent := new(martcconf.RTCEvent)
				newEvent.Messages = nil
				newEvent.Title = oTitle
				newEvent.Oneshot = rxPacket.Payloads[7][0] == '1'

				hexDecode, decodeError = hex.DecodeString(libmmp.Payload_2_str(rxPacket.Payloads[8]))
				if (decodeError != nil) {

					libmarty.Marty_full_log("Couldn't decode day mask scheduling '%s' at %s!\n",
						libmarty.LOG_ERROR, &data, oTitle, oTime)
					applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["EXECUTE"]
					goto JMP_TX_STATUS
				}

				newEvent.DayMask = hexDecode[0]

				hexDecode, decodeError = hex.DecodeString(libmmp.Payload_2_str(rxPacket.Payloads[9]))
				if (decodeError != nil || len(hexDecode) != 2) {

					libmarty.Marty_full_log("Couldn't decode month mask scheduling '%s' at %s!\n",
						libmarty.LOG_ERROR, &data, oTitle, oTime)
					applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["EXECUTE"]
					goto JMP_TX_STATUS
				}

				newEvent.MonthMask = uint16(hexDecode[0]) << 8 | uint16(hexDecode[1])
				newEvent.Skip = rxPacket.Payloads[10][0] == '1'

				martcconf.MaRTCConf[oTime] = append(martcconf.MaRTCConf[oTime], *newEvent)

				libmarty.Marty_full_log("Created a new event '%s' at %s.\n",
					libmarty.LOG_SUCCESS, &data, oTitle, oTime)

			case mendpoints.MARTSS_EPT_APPEND:

				if tickEvent == nil {

					libmarty.Marty_full_log("No event '%s' at %s!\n",
						libmarty.LOG_ERROR, &data, oTitle, oTime)
					applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["EXECUTE"]
					goto JMP_TX_STATUS
				}

				eventHandle := new(libmarty.MartyMessageHandle)
				libmarty.Marty_message_handle_factory(eventHandle)

				eventHandle.Execute = new(libmmp.MMPacket)
				eventHandle.Execute.Type = libmmp.MMP_TYPE_SEND
				eventHandle.Execute.Argument = libmmp.MMPArgs[libmmp.MMP_TYPE_SEND]["DIRECT"]
				eventHandle.Execute.Payloads = libmmp.Payloads_2_heap(rxPacket.Payloads[7:len(rxPacket.Payloads)])

				tickEvent.Messages = append(tickEvent.Messages, *eventHandle)

				libmarty.Marty_full_log("Appended a new handle to '%s' at %s.\n",
					libmarty.LOG_SUCCESS, &data, oTitle, oTime)

			case mendpoints.MARTSS_EPT_SKIP:

				if tickEvent == nil {

					libmarty.Marty_full_log("No event '%s' at %s!\n",
						libmarty.LOG_ERROR, &data, oTitle, oTime)
					applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["EXECUTE"]
					goto JMP_TX_STATUS
				}
				tickEvent.Skip = true

				libmarty.Marty_full_log("Registered skip for '%s' at %s.\n",
					libmarty.LOG_SUCCESS, &data, oTitle, oTime)

			case mendpoints.MARTSS_EPT_RESCHEDULE:

				if tickEvent == nil {

					libmarty.Marty_full_log("No event '%s' at %s!\n",
						libmarty.LOG_ERROR, &data, oTitle, oTime)
					applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["EXECUTE"]
					goto JMP_TX_STATUS
				}

				if event_exists(oTitle, nTime) {

					libmarty.Marty_full_log("Event '%s' at %s already exists! Won't schedule.\n",
						libmarty.LOG_ERROR, &data, oTitle, nTime)
					applicationStatus = libmmp.MMPArgs[libmmp.MMP_TYPE_APPLICATION]["EXECUTE"]
					goto JMP_TX_STATUS
				}

				tickEvent.Skip = true
				nTime = libmmp.Payload_2_str(rxPacket.Payloads[7])
				tickEvents, mapLookup = martcconf.MaRTCConf[nTime]

				if (!mapLookup) {
					martcconf.MaRTCConf[nTime] = nil
					tickEvents = martcconf.MaRTCConf[nTime]
				}

				tickEvent = martcconf.RTC_2_heap(*tickEvent)
				tickEvent.Skip = false
				tickEvent.Oneshot = true

				martcconf.MaRTCConf[nTime] = append(tickEvents, *tickEvent)

				libmarty.Marty_full_log("Rescheduled '%s' at %s to %s.\n",
					libmarty.LOG_SUCCESS, &data, oTitle, oTime, nTime)

		}

		JMP_TX_STATUS:

		libmarty.Marty_set_application(&txPacket, &status,
			rxPacket.Payloads[0], applicationStatus)

		if (status.Error != libmmp.NONE) {
			libmarty.Marty_full_log("Couldn't set autopacket!\n", libmarty.LOG_ERROR, &data)
			continue
		}

		libmmp.Serialize(&txPacket, &status)
		if (status.Error != libmmp.NONE) {
			libmarty.Marty_full_log("Couldn't serialize autopacket!\n", libmarty.LOG_ERROR, &data)
			fmt.Printf("%d\n", status.Error)
			continue
		}

		libmarty.Marty_net_write(&txPacket, &status, &data)
	}
}
